﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.IO;
using System.Globalization;
using System.Diagnostics;
using System.Threading;
using Accord.Math.Decompositions;

namespace CADemo
{
    public partial class CADemoMain : Form
    {
        // Minimum Mass scale
        private const float _cminscale = 0.5f;

        // Data table
        private DataTable _data = null;
        // Round precision
        private int _precision = 6;
        // Max and Min coordinates for rows in the two dimensions
        private SizeF _rowdims1 = SizeF.Empty;
        private SizeF _rowdims2 = SizeF.Empty;
        private SizeF _rowdimp1 = SizeF.Empty;
        private SizeF _rowdimp2 = SizeF.Empty;
        // Max and Min coordinates for columns in the two dimensions
        private SizeF _coldims1 = SizeF.Empty;
        private SizeF _coldims2 = SizeF.Empty;
        private SizeF _coldimp1 = SizeF.Empty;
        private SizeF _coldimp2 = SizeF.Empty;
        // Max and Min coordinates for the graph axis
        private SizeF _sizex = SizeF.Empty;
        private SizeF _sizey = SizeF.Empty;
        // Mass scale for rows and columns
        private float _rmasssc = 0f;
        private float _cmasssc = 0f;
        private float _minrmass = 0f;
        private float _mincmass = 0f;

        public CADemoMain()
        {
            InitializeComponent();
            dgData.AutoGenerateColumns = true;
            tbWizzard.TabPages.Remove(tpGraph);
        }

        private void SetGraphScales()
        {
            // Min and Max coordinates
            if (bSymm.Checked)
            {
                _sizex = new SizeF(Math.Min(_rowdimp1.Width, _coldimp1.Width), Math.Max(_rowdimp1.Height, _coldimp1.Height));
                _sizey = new SizeF(Math.Min(_rowdimp2.Width, _coldimp2.Width), Math.Max(_rowdimp2.Height, _coldimp2.Height));
            }
            else if (bRow.Checked)
            {
                _sizex = new SizeF(Math.Min(_rowdimp1.Width, _coldims1.Width), Math.Max(_rowdimp1.Height, _coldims1.Height));
                _sizey = new SizeF(Math.Min(_rowdimp2.Width, _coldims2.Width), Math.Max(_rowdimp2.Height, _coldims2.Height));
            }
            else
            {
                _sizex = new SizeF(Math.Min(_rowdims1.Width, _coldimp1.Width), Math.Max(_rowdims1.Height, _coldimp1.Height));
                _sizey = new SizeF(Math.Min(_rowdims2.Width, _coldimp2.Width), Math.Max(_rowdims2.Height, _coldimp2.Height));
            }
        }

        private float RowCoordinate(int dim, int row)
        {
            if (bRow.Checked || bSymm.Checked)
            {
                // Ros in principal coordinates
                return dim == 0 ? Convert.ToSingle(_data.Rows[row][_data.Columns.Count - 2]) : Convert.ToSingle(_data.Rows[row][_data.Columns.Count - 1]);
            }
            else 
            {
                // Rows in standard coordinates
                return dim == 0 ? Convert.ToSingle(_data.Rows[row][_data.Columns.Count - 4]) : Convert.ToSingle(_data.Rows[row][_data.Columns.Count - 3]);
            }
        }

        private float ColCoordinate(int dim, int col)
        {
            if (bCol.Checked || bSymm.Checked)
            {
                // Columns in principal coordiantes
                return dim == 0 ? Convert.ToSingle(_data.Rows[_data.Rows.Count - 2][col]) : Convert.ToSingle(_data.Rows[_data.Rows.Count - 1][col]);
            }
            else
            {
                // Columns in standard coordinates
                return dim == 0 ? Convert.ToSingle(_data.Rows[_data.Rows.Count - 4][col]) : Convert.ToSingle(_data.Rows[_data.Rows.Count - 3][col]);
            }
        }

        private float RowMass(int row, float max)
        {
            return max * (bMass.Checked ? _cminscale + (Convert.ToSingle(_data.Rows[row][_data.Columns.Count - 7]) - _minrmass) * _rmasssc : 1f);
        }

        private float ColMass(int col, float max)
        {
            return max * (bMass.Checked ? _cminscale + (Convert.ToSingle(_data.Rows[_data.Rows.Count - 7][col]) - _mincmass) * _cmasssc : 1f);
        }

        private void bOpen_Click(object sender, EventArgs e)
        {
            if (OFDlg.ShowDialog() == DialogResult.OK)
            {
                StreamReader rdr = new StreamReader(OFDlg.FileName);
                bsData.DataSource = null;
                bChid.Enabled = false;
                bProject2D.Enabled = false;
                tbWizzard.TabPages.Remove(tpGraph);
                int.TryParse(txtPrec.Text, out _precision);
                if ((_precision <= 0) || (_precision > 16))
                {
                    _precision = 6;
                    txtPrec.Text = "6";
                }
                try
                {
                    string line = rdr.ReadLine();
                    if (line != null)
                    {
                        // Add Columns
                        string[] columns = line.Split(txtSep.Text[0]);
                        _data = new DataTable();
                        // First Column for Row Names
                        _data.Columns.Add(new DataColumn("row/col", typeof(string)));
                        for (int ix = 1; ix < columns.Length; ix++)
                        {
                            _data.Columns.Add(new DataColumn(columns[ix].Replace("\"", ""), typeof(double)));
                        }
                        // Add Row Mass Column
                        _data.Columns.Add(new DataColumn("Mass", typeof(double)));
                        DataRow row;
                        double total = 0;
                        // Min and Max Mass value
                        double max = double.MinValue;
                        double min = double.MaxValue;

                        while ((line = rdr.ReadLine()) != null)
                        {
                            // Add Rows
                            row = _data.NewRow();
                            string[] rowdata = line.Split(txtSep.Text[0]);
                            // First Element is the Name
                            row[0] = rowdata[0].Replace("\"", "");
                            double rtotal = 0;
                            for (int ix = 1; ix < rowdata.Length; ix++)
                            {
                                // Add Values
                                double v = Convert.ToDouble(rowdata[ix].Replace(txtDecSep.Text, CultureInfo.CurrentCulture.NumberFormat.NumberDecimalSeparator));
                                total += v;
                                rtotal += v;
                                row[ix] = Math.Round(v, _precision);
                            }
                            row[_data.Columns.Count - 1] = Math.Round(rtotal, _precision);
                            max = Math.Max(rtotal, max);
                            min = Math.Min(rtotal, min);
                            _data.Rows.Add(row);
                        }
                        // Row Mass scale
                        _minrmass = (float)min;
                        _rmasssc = (float)((1 - _cminscale) / (max - min));

                        max = double.MinValue;
                        min = double.MaxValue;
                        // Add Column Mass Row
                        row = _data.NewRow();
                        row[0] = "Mass";                        
                        for (int ix = 1; ix < _data.Columns.Count - 1; ix++)
                        {
                            double ctotal = 0;
                            for (int ir = 0; ir < _data.Rows.Count; ir++)
                            {
                                ctotal += (double)_data.Rows[ir][ix];
                            }
                            row[ix] = Math.Round(ctotal, _precision);
                            max = Math.Max(ctotal, max);
                            min = Math.Min(ctotal, min);
                        }
                        row[_data.Columns.Count - 1] = Math.Round(total, _precision);
                        _data.Rows.Add(row);
                        // Column Mass scale
                        _mincmass = (float)min;
                        _cmasssc = (float)((1 - _cminscale) / (max - min));

                        // Bind Table Data to Grid
                        bsData.DataSource = _data;
                        bChid.Enabled = true;
                    }
                    else
                    {
                        throw new Exception("Empty file");
                    }
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message);
                }
                finally
                {
                    if (rdr != null)
                    {
                        rdr.Close();
                    }
                }
            }
        }

        private void bChid_Click(object sender, EventArgs e)
        {
            if (_data != null)
            {
                try
                {                    
                    // Total inertia
                    double inertia = 0;
                    _data.Columns.Add("Chi-Dist", typeof(double));
                    _data.Columns.Add("Inertia", typeof(double));
                    // Calculate for the rows
                    for (int ir = 0; ir < _data.Rows.Count - 1; ir++)
                    {
                        double chdrow = 0;
                        for (int ic = 1; ic < _data.Columns.Count - 3; ic++)
                        {
                            // (observed (row profile) - expected)^2 / expected
                            chdrow += Math.Pow(((double)_data.Rows[ir][ic] / (double)_data.Rows[ir][_data.Columns.Count - 3]) - (double)_data.Rows[_data.Rows.Count - 1][ic], 2) / (double)_data.Rows[_data.Rows.Count - 1][ic];
                        }
                        // Chi-distance
                        _data.Rows[ir][_data.Columns.Count - 2] = Math.Round(Math.Sqrt(chdrow), _precision);
                        // Inertia
                        inertia += chdrow * (double)_data.Rows[ir][_data.Columns.Count - 3];
                        _data.Rows[ir][_data.Columns.Count - 1] = Math.Round(chdrow * (double)_data.Rows[ir][_data.Columns.Count - 3], _precision);
                    }
                    DataRow rowc = _data.NewRow();
                    rowc[0] = "Chi-Dist";
                    DataRow rowi = _data.NewRow();
                    rowi[0] = "Inertia";
                    // Calculate for the columns
                    for (int ic = 1; ic < _data.Columns.Count - 3; ic++)
                    {
                        double chdcol = 0;
                        for (int ir = 0; ir < _data.Rows.Count - 1; ir++)
                        {
                            // (observed (column profile) - expected)^2 / expected
                            chdcol += Math.Pow(((double)_data.Rows[ir][ic] / (double)_data.Rows[_data.Rows.Count - 1][ic]) - (double)_data.Rows[ir][_data.Columns.Count - 3], 2) / (double)_data.Rows[ir][_data.Columns.Count - 3];
                        }
                        // Chi-distance
                        rowc[ic] = Math.Round(Math.Sqrt(chdcol), _precision);
                        // Inertia
                        rowi[ic] = Math.Round(chdcol * (double)_data.Rows[_data.Rows.Count - 1][ic], _precision);
                    }
                    // Total inertia
                    rowi[_data.Columns.Count - 1] = Math.Round(inertia, _precision);
                    _data.Rows.Add(rowc);
                    _data.Rows.Add(rowi);
                    bsData.DataSource = _data;
                    bChid.Enabled = false;
                    bProject2D.Enabled = true;
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message);
                }
            }
        }

        private void bProject2D_Click(object sender, EventArgs e)
        {
            if (_data != null)
            {
                try
                {
                    // Matrix S = (P - rc(t)) / sqrt(rc), where P is the contingency table, r is the mass of rows and c the mass of columns in diagonal form
                    double[,] wdata = new double[_data.Rows.Count - 3, _data.Columns.Count - 4];
                    for (int ir = 0; ir < _data.Rows.Count - 3; ir++)
                    {
                        for (int ic = 1; ic < _data.Columns.Count - 3; ic++)
                        {
                            wdata[ir, ic - 1] = Math.Round(((double)_data.Rows[ir][ic] - (double)_data.Rows[ir][_data.Columns.Count - 3] * (double)_data.Rows[_data.Rows.Count - 3][ic]) / 
                                Math.Sqrt((double)_data.Rows[ir][_data.Columns.Count - 3] * (double)_data.Rows[_data.Rows.Count - 3][ic]), _precision);
                        }
                    }
                    // Singula value decomposition
                    SingularValueDecomposition svd = new SingularValueDecomposition(wdata);
                    // Standard coordinates
                    _data.Columns.Add("Dim1 S.", typeof(double));
                    _data.Columns.Add("Dim2 S.", typeof(double));
                    // Principal coordinates
                    _data.Columns.Add("Dim1 P.", typeof(double));
                    _data.Columns.Add("Dim2 P.", typeof(double));
                    // Processing rows
                    float[,] vmax = new float[2, 4]; 

                    for (int ir = 0; ir < _data.Rows.Count - 3; ir++)
                    {
                        for (int iu = 0; iu < 2; iu++)
                        {
                            // Project using U vectors of SVD divided by the sqrt of mass
                            double v = svd.LeftSingularVectors[ir, iu] / Math.Sqrt((double)_data.Rows[ir][_data.Columns.Count - 7]);
                            // Standard coordinates
                            _data.Rows[ir][iu + (_data.Columns.Count - 4)] = Math.Round(v, _precision);
                            // Principal coordinates using singular value
                            _data.Rows[ir][iu + (_data.Columns.Count - 2)] = Math.Round(v * svd.Diagonal[iu], _precision);
                            // Store max and min coordinates for each row & dimension
                            vmax[iu, 0] = (float)Math.Min(vmax[iu, 0], Math.Round(v, _precision));
                            vmax[iu, 1] = (float)Math.Max(vmax[iu, 1], Math.Round(v, _precision));
                            vmax[iu, 2] = (float)Math.Min(vmax[iu, 2], Math.Round(v * svd.Diagonal[iu], _precision));
                            vmax[iu, 3] = (float)Math.Max(vmax[iu, 3], Math.Round(v * svd.Diagonal[iu], _precision));
                        }
                    }

                    _rowdims1 = new SizeF(vmax[0, 0], vmax[0, 1]);
                    _rowdims2 = new SizeF(vmax[1, 0], vmax[1, 1]);
                    _rowdimp1 = new SizeF(vmax[0, 2], vmax[0, 3]);
                    _rowdimp2 = new SizeF(vmax[1, 2], vmax[1, 3]);

                    // Same for columns
                    DataRow row1s = _data.NewRow();
                    row1s[0] = "Dim1 S.";
                    DataRow row2s = _data.NewRow();
                    row2s[0] = "Dim2 S.";
                    DataRow row1p = _data.NewRow();
                    row1p[0] = "Dim1 P.";
                    DataRow row2p = _data.NewRow();
                    row2p[0] = "Dim2 P.";

                    vmax = new float[2, 4]; 

                    for (int ic = 1; ic < _data.Columns.Count - 7; ic++)
                    {
                        double v1 = svd.RightSingularVectors[ic - 1, 0] / Math.Sqrt((double)_data.Rows[_data.Rows.Count - 3][ic]);
                        row1s[ic] = Math.Round(v1, _precision);
                        row1p[ic] = Math.Round(v1 * svd.Diagonal[0], _precision);
                        vmax[0, 0] = (float)Math.Min(vmax[0, 0], Math.Round(v1, _precision));
                        vmax[0, 1] = (float)Math.Max(vmax[0, 1], Math.Round(v1, _precision));
                        vmax[0, 2] = (float)Math.Min(vmax[0, 2], Math.Round(v1 * svd.Diagonal[0], _precision));
                        vmax[0, 3] = (float)Math.Max(vmax[0, 3], Math.Round(v1 * svd.Diagonal[0], _precision));
                        double v2 = svd.RightSingularVectors[ic - 1, 1] / Math.Sqrt((double)_data.Rows[_data.Rows.Count - 3][ic]);
                        row2s[ic] = Math.Round(v2, _precision);
                        row2p[ic] = Math.Round(v2 * svd.Diagonal[1], _precision);
                        vmax[1, 0] = (float)Math.Min(vmax[1, 0], Math.Round(v2, _precision));
                        vmax[1, 1] = (float)Math.Max(vmax[1, 1], Math.Round(v2, _precision));
                        vmax[1, 2] = (float)Math.Min(vmax[1, 2], Math.Round(v2 * svd.Diagonal[1], _precision));
                        vmax[1, 3] = (float)Math.Max(vmax[1, 3], Math.Round(v2 * svd.Diagonal[1], _precision));
                    }

                    _coldims1 = new SizeF(vmax[0, 0], vmax[0, 1]);
                    _coldims2 = new SizeF(vmax[1, 0], vmax[1, 1]);
                    _coldimp1 = new SizeF(vmax[0, 2], vmax[0, 3]);
                    _coldimp2 = new SizeF(vmax[1, 2], vmax[1, 3]);

                    _data.Rows.Add(row1s);
                    _data.Rows.Add(row2s);
                    _data.Rows.Add(row1p);
                    _data.Rows.Add(row2p);
                    // Calculate total inertias for each dimension (row standard coordinates * contingency table * column standard coordinates)
                    for (int iu = 0; iu < 2; iu++)
                    {
                        double itotal = 0;
                        for (int ic = 1; ic < _data.Columns.Count - 7; ic++)
                        {
                            double rtmp = 0;
                            for (int ir = 0; ir < _data.Rows.Count - 7; ir++)
                            {
                                rtmp += (double)_data.Rows[ir][ic] * (double)_data.Rows[ir][iu + (_data.Columns.Count - 4)];
                            }
                            itotal += rtmp * (double)_data.Rows[iu + (_data.Rows.Count - 4)][ic];
                        }
                        // Put in the inertia row and column
                        _data.Rows[_data.Rows.Count - 5][iu + (_data.Columns.Count - 4)] = Math.Round(itotal * itotal, _precision);
                        _data.Rows[iu + (_data.Rows.Count - 4)][_data.Columns.Count - 5] = Math.Round(itotal * itotal, _precision);
                    }
                    bsData.DataSource = _data;
                    bProject2D.Enabled = false;
                    SetGraphScales();
                    tbWizzard.TabPages.Add(tpGraph);                    
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message);
                }
            }
        }

        private void pGraph_Paint(object sender, PaintEventArgs e)
        {
            if (_data != null)
            {
                e.Graphics.FillRectangle(Brushes.White, pGraph.ClientRectangle);
                // Draw the axis, x for dimension 1 and y for dimension 2
                Font f = new Font("Arial", 10f, GraphicsUnit.Pixel);

                float offx = 20f + Math.Max(e.Graphics.MeasureString(_sizex.Width.ToString(), f).Width, e.Graphics.MeasureString(_sizex.Height.ToString(), f).Width);
                float offy = 20f + Math.Max(e.Graphics.MeasureString(_sizey.Width.ToString(), f).Width, e.Graphics.MeasureString(_sizey.Height.ToString(), f).Width);

                float wgraph = _sizex.Height - _sizex.Width;
                float hgraph = _sizey.Height - _sizey.Width;

                float rangex = (float)pGraph.Width - (10f + offx);
                float rangey = (float)pGraph.Height - (10f + offy);

                // Scale the size of the graph to the window
                if ((rangex * (hgraph / wgraph)) <= rangey)
                {
                    rangey = rangex * (hgraph / wgraph);
                }
                else
                {
                    rangex = rangey * (wgraph / hgraph);
                }

                float scalex = rangex / wgraph;
                float scaley = rangey / hgraph;

                Matrix oldm = e.Graphics.Transform;
                e.Graphics.TranslateTransform(0f, (float)pGraph.Height);
                e.Graphics.ScaleTransform(1f, -1f);

                // Draw the zero lines

                Pen pz = new Pen(Color.LightGray);
                pz.DashStyle = DashStyle.DashDot;

                e.Graphics.DrawLine(pz, offx + (scalex * -_sizex.Width), offy, offx + (scalex * -_sizex.Width), offy + rangey);
                e.Graphics.DrawLine(pz, offx, offy + (scaley * -_sizey.Width), offx + rangex, offy + (scaley * -_sizey.Width));

                pz.Dispose();

                // Draw axis lines

                e.Graphics.DrawLine(Pens.Black, offx, offy, offx, offy + rangey);
                e.Graphics.DrawLine(Pens.Black, offx, offy, offx + rangex, offy);

                // Draw the axis scales

                float cstep = hgraph / 5f;
                float axisstep = rangey / 5f;

                e.Graphics.ResetTransform();

                for (int y = 0; y < 6; y++)
                {
                    e.Graphics.DrawString((_sizey.Width + (float)y * cstep).ToString(), f, Brushes.Black, 0f, ((float)pGraph.Height - offy) - 5f - (float)y * axisstep);
                    e.Graphics.DrawLine(Pens.Black, offx - 5f, ((float)pGraph.Height - offy) - (float)y * axisstep, offx, ((float)pGraph.Height - offy) - (float)y * axisstep);
                }

                cstep = wgraph / 5f;
                axisstep = rangex / 5f;

                e.Graphics.TranslateTransform(offx - 5f, (float)pGraph.Height);
                e.Graphics.RotateTransform(-90f);

                for (int x = 0; x < 6; x++)
                {
                    e.Graphics.DrawString((_sizex.Width + (float)x * cstep).ToString(), f, Brushes.Black, 0f, (float)x * axisstep);
                    e.Graphics.DrawLine(Pens.Black, offy - 5f, 5f + (float)x * axisstep, offy, 5f + (float)x * axisstep);
                }                

                e.Graphics.ResetTransform();
                e.Graphics.TranslateTransform(offx + (scalex * -_sizex.Width), (float)pGraph.Height - offy - (scaley * -_sizey.Width));
                e.Graphics.ScaleTransform(1f, -1f);

                // Draw the points for the rows

                for (int ir = 0; ir < _data.Rows.Count - 7; ir++)
                {
                    float x = RowCoordinate(0, ir);
                    float y = RowCoordinate(1, ir);
                    float w = RowMass(ir, 16f);

                    e.Graphics.FillEllipse(Brushes.Blue, new RectangleF(scalex * x - w / 2f, scaley * y - w / 2f, w, w));
                }

                // Draw the points for the columns

                for (int ic = 1; ic < _data.Columns.Count - 7; ic++)
                {
                    float x = ColCoordinate(0, ic);
                    float y = ColCoordinate(1, ic);
                    float w = ColMass(ic, 16f);

                    e.Graphics.FillEllipse(Brushes.Red, new RectangleF(scalex * x - w / 2f, scaley * y - w / 2f, w, w));
                }

                e.Graphics.ResetTransform();

                // Draw the labels for the rows

                for (int ir = 0; ir < _data.Rows.Count - 7; ir++)
                {
                    float x = RowCoordinate(0, ir) - _sizex.Width;
                    float y = RowCoordinate(1, ir) - _sizey.Width;
                    float w = RowMass(ir, 16f);

                    e.Graphics.DrawString(_data.Rows[ir][0].ToString(), f, Brushes.Black, offx + 1f + (w / 2f) + scalex * x, ((float)pGraph.Height - offy - (w / 2f)) - scaley * y);
                }

                // Draw the labels for the columns

                for (int ic = 1; ic < _data.Columns.Count - 7; ic++)
                {
                    float x = ColCoordinate(0, ic) - _sizex.Width;
                    float y = ColCoordinate(1, ic) - _sizey.Width;
                    float w = ColMass(ic, 16f);

                    e.Graphics.DrawString(_data.Columns[ic].ColumnName, f, Brushes.Black, offx + 1f + (w / 2f) + scalex * x, ((float)pGraph.Height - offy - (w / 2f)) - scaley * y);
                }

                f.Dispose();
            }
        }

        private void bCol_Click(object sender, EventArgs e)
        {
            bCol.Checked = true;
            bRow.Checked = false;
            bSymm.Checked = false;
            SetGraphScales();
            pGraph.Invalidate();
        }

        private void bRow_Click(object sender, EventArgs e)
        {
            bCol.Checked = false;
            bRow.Checked = true;
            bSymm.Checked = false;
            SetGraphScales();
            pGraph.Invalidate();
        }

        private void bSymm_Click(object sender, EventArgs e)
        {
            bCol.Checked = false;
            bRow.Checked = false;
            bSymm.Checked = true;
            SetGraphScales();
            pGraph.Invalidate();
        }

        private void bMass_Click(object sender, EventArgs e)
        {
            bMass.Checked = !bMass.Checked;
            pGraph.Invalidate();
        }

        private void bHelp_Click(object sender, EventArgs e)
        {
            Process p = new Process();
            string culture = "en";
            switch (Thread.CurrentThread.CurrentCulture.TwoLetterISOLanguageName)
            {
                case "es":
                    culture = "es";
                    break;
                case "an":  // Aragones
                    culture = "es";
                    break;
                case "ca":  // Catalán
                    culture = "es";
                    break;
                case "eu":  // Euskera
                    culture = "es";
                    break;
                case "gl":  // gallego
                    culture = "es";
                    break;
            }
            if (culture == "en")
            {
                p.StartInfo.FileName = "http://software-tecnico-libre.es/en/article-by-topic/software-developing/charp-programming/mathematics-with-csharp/correspondence-analysis-csharp";
            }
            else
            {
                p.StartInfo.FileName = "http://software-tecnico-libre.es/es/articulo-por-tema/desarrollo-de-software/programacion-en-csharp/matematicas-con-csharp/analisis-de-correspondencias-con-csharp";
            }
            p.Start();
        }
    }
}
